import { Archetype, Query } from "./Archetype";
import { WithRequiredKeys } from "./util/types";
export declare type EntityId = number;
/**
 * Entities in Miniplex are just plain old Javascript objects. We are assuming
 * map-like objects that use string-based properties to identify individual components.
 */
export interface IEntity {
    [key: string]: ComponentData;
}
/**
 * Miniplex uses an internal component that it will automatically add to all created
 * entities.
 */
export declare type MiniplexComponent<T> = {
    __miniplex: {
        id: EntityId;
        world: World<T>;
        archetypes: Archetype<T>[];
    };
};
export declare type RegisteredEntity<T extends IEntity> = T & MiniplexComponent<T>;
/**
 * For situations where no entity type argument is passed to createWorld, we'll
 * default to an untyped entity type that can hold any component.
 */
export declare type UntypedEntity = {
    [components: string]: ComponentData;
};
/**
 * Component names are just strings/object property names.
 */
export declare type ComponentName<T extends IEntity> = keyof T;
/**
 * The data of a component can be literally anything. Good times!
 */
export declare type ComponentData = any;
/**
 * Utility type that represents an Entity that is guaranteed to have the specified
 * component(s) available.
 */
export declare type EntityWith<TEntity extends IEntity, TComponents extends ComponentName<TEntity>> = WithRequiredKeys<TEntity, TComponents> & TEntity;
/**
 * A tag is just an "empty" component. For convenience and nicer type support, we're
 * providing a Tag type and constant, both of which are simply equal to `true`.
 */
export declare const Tag = true;
export declare type Tag = true;
export declare class World<T extends IEntity = UntypedEntity> {
    /** An array holding all entities known to this world. */
    entities: (RegisteredEntity<T> | null)[];
    /** A list of known archetypes. */
    private archetypes;
    archetype<TQuery extends Query<T>>(...query: TQuery): Archetype<T, TQuery>;
    private indexEntity;
    private unregisterEntity;
    createEntity<NewEntity extends T>(entity: NewEntity & T): NewEntity & MiniplexComponent<T>;
    createEntity(...partials: Partial<T>[]): RegisteredEntity<T>;
    destroyEntity: (entity: RegisteredEntity<T>) => void;
    addComponent: (entity: RegisteredEntity<T>, ...partials: Partial<T>[]) => void;
    removeComponent: (entity: RegisteredEntity<T>, ...components: ComponentName<T>[]) => void;
    /**
     * A queue of commands to run.
     */
    private queuedCommands;
    queue: {
        createEntity: (entity: T) => void;
        destroyEntity: (entity: RegisteredEntity<T> | T) => void;
        addComponent: (entity: RegisteredEntity<T>, ...partials: Partial<T>[]) => void;
        removeComponent: (entity: RegisteredEntity<T>, ...names: ComponentName<T>[]) => void;
        flush: () => void;
    };
    /**
     * Clear the entire world by discarding all entities and indices.
     */
    clear(): void;
}
